/*------------------------------------------------------------------------------

                          DATA BASE BROWSE EXAMPLE

  Common data base browsing system of Visual Clipper Ver. 1.0

  (c) 1993 - Francisco Morero Peyrona

------------------------------------------------------------------------------*/

#include "inkey.ch"
#include "setcurs.ch"
#include "Designer.ch"

static lAppend := .F.

func BrowseDbf()

local oB, oCol, n, nWnd

local nCurSave := SetCursor( 0 )
local nTop     := 3,;
      nLeft    := 1,;
      nBottom  := nMAXROW-2,;
      nRight   := nMAXCOL-3

   // create a TBrowse object for a database
   oB := TBrowseDB( nTop + 1, nLeft + 1, nBottom - 2, nRight - 3 )
   oB:headSep   = ""
   oB:colorSpec = aClr[ CLR_DLGBOX, 4 ] + "," + aClr[ CLR_GET ]
   oB:skipBlock := { |x| Skipped( x, lAppend ) }

   // add one column for each field
   for n := 1 to Fcount()
       oCol = TBColumnNew( FieldName(n), FieldBlock( FieldName(n) ) )
       oCol:footSep = ""
       oB:addColumn( oCol )
   next

   @ nTop, nLeft, nBottom, nRight CREATE WINDOW nWnd          ;
     TITLE "Browsing <" + Alias() + ">"                       ;
     CONTROL oB  READER { |nWnd, oB| _BrowseDbf( nWnd, oB ) }

   @ nBottom - 1, nLeft+3 SAY "Record" COLOR aClr[ CLR_DLGBOX, 4 ]

   ACTIVATE WINDOW nWnd KILL

   SetCursor( nCurSave )

return NIL

//----------------------------------------------------------------------------//

static func _BrowseDbf( nWnd, oB )

   local lKillAppend, nKey, lGotKey, bKeyBlock
   local lMore := .T.

   if ( Eof() )
      DbGoTop()
   end

   // init
   lAppend := lKillAppend := .F.
   RtnCursor( .F. )
   while ( !oB:stabilize() ) ; end
   RtnCursor( .T. )

   if ( LastRec() == 0 )
      // empty file..force append mode
      nKey := K_DOWN
      lGotKey := .t.
   else
      lGotKey := .f.
   end

   while (lMore)
      if ( !lGotKey )
         RtnCursor( .F. )
         while ( !oB:stabilize() )
            if ( (nKey := Inkey()) != 0 )
                    lGotKey := .t.
                    exit
            end
         end
         RtnCursor( .T. )
      end

      WINDOW nWnd BARSCROLL VERTICAL   RecNo()      OF LastRec()

      WINDOW nWnd BARSCROLL HORIZONTAL oB:ColPos()  OF oB:ColCount()

      if ( !lGotKey )
              // the TBrowse object is stable
         if ( oB:hitBottom )
            if ( !lAppend .or. Recno() != LastRec() + 1 )
               if ( lAppend )
                       // continue appending..restore color to current row
                       oB:refreshCurrent()
                       RtnCursor( .F. )
                       while ( !oB:stabilize() ) ; end
                       RtnCursor( .T. )

                       // ensure bottom of file without refresh
                       go bottom
               else
                       // begin append mode
                       lAppend := .t.

                       // turn the cursor on
                       SetCursor( if(ReadInsert(), SC_INSERT, SC_NORMAL) )
               end

               // move to next row and stabilize to set rowPos
               oB:down()
               RtnCursor( .F. )
               while ( !oB:stabilize() ) ; end
               RtnCursor( .T. )

               // color the row
               oB:colorRect({oB:rowPos,1,oB:rowPos,oB:colCount},{2,2})
            end
         end

         // display status and stabilize again for correct cursor pos
         Statline(oB, lAppend)
         RtnCursor( .F. )
         while ( !oB:stabilize() ) ; end
         RtnCursor( .T. )

         // idle
         nKey := Inkey(0)

         if ( (bKeyBlock := SetKey(nKey)) != NIL )
            // run SET KEY block
            Eval(bKeyBlock, ProcName(1), ProcLine(1), "")
            loop
         end
      else
        // reset for next loop
        lGotKey := .f.
      end

      do case
      case ( nKey == K_DOWN )
              if ( lAppend )
                 oB:hitBottom := .t.
              else
                 oB:down()
              end

      case ( nKey == K_UP )
              if ( lAppend )
                      lKillAppend := .t.
              else
                      oB:up()
              end

      case ( nKey == K_PGDN )
              if ( lAppend )
                      oB:hitBottom := .t.
              else
                      oB:pageDown()
              end

      case ( nKey == K_PGUP )
              if ( lAppend )
                      lKillAppend := .t.
              else
                      oB:pageUp()
              end

      case ( nKey == K_CTRL_PGUP )
              if ( lAppend )
                      lKillAppend := .t.
              else
                      oB:goTop()
              end

      case ( nKey == K_CTRL_PGDN )
              if ( lAppend )
                      lKillAppend := .t.
              else
                      oB:goBottom()
              end

      case ( nKey == K_RIGHT )           ; oB:right()
      case ( nKey == K_LEFT )            ; oB:left()
      case ( nKey == K_HOME )            ; oB:home()
      case ( nKey == K_END )             ; oB:end()
      case ( nKey == K_CTRL_LEFT )       ; oB:panLeft()
      case ( nKey == K_CTRL_RIGHT )      ; oB:panRight()
      case ( nKey == K_CTRL_HOME )       ; oB:panHome()
      case ( nKey == K_CTRL_END )        ; oB:panEnd()

      case ( nKey == K_INS )
              // begin append mode
              DbGoBottom()
              nKey := K_DOWN
              lGotKey := .t.

      case ( nKey == K_DEL )
              // delete key..toggle deleted() flag
              if ( Recno() != LastRec() + 1 )
                 if ( Deleted() )
                    RECALL
                 else
                    DELETE
                 end
              end

      case ( nKey == K_RETURN )
              // edit
              if ( lAppend .or. Recno() != LastRec() + 1 )
                      nKey := DoGet(oB, lAppend)

                      // use returned value as next key if not zero
                      lGotKey := ( nKey != 0 )
              else
                      // begin append mode
                      nKey := K_DOWN
                      lGotKey := .t.
              end

      case ( nKey == K_ESC )
              // exit browse
              lMore := .f.

      otherwise
          if ( nKey >= 32 .and. nKey <= 255 )
             // begin edit and supply the first character
             KEYBOARD Chr( K_RETURN ) + Chr( nKey )
          end
      end

      if ( lKillAppend )
         // turn off append mode
         lKillAppend := .f.
         lAppend := .f.

         // refresh respecting any change in index order
         FreshOrder(oB)
         SetCursor(0)
      end
   end

return NIL


/***
*	DoGet()
*
*	Edit the current field
*/

static func DoGet(oB, lAppend)

local lExitSave
local oCol, oGet, nKey, cExpr, xEval
local lFresh, nCursSave, mGetVar
local lRtnCrs := RtnCursor( .F. )
local lRtnProc:= RtnSetProc( .F. )

   // make sure the display is correct
   oB:hitTop := .f.
   Statline(oB, lAppend)
   while ( !oB:stabilize() ) ; end
   RtnCursor( .T. )

   // save state
   lExitSave := Set(_SET_EXIT, .t.)
   nCursSave := SetCursor( if(ReadInsert(), SC_INSERT, SC_NORMAL) )

   // get the controlling index key
   cExpr := IndexKey(0)
   if ( !Empty(cExpr) )
      // expand key expression for later comparison
      xEval := &cExpr
   end

   // get column object from browse
   oCol := oB:getColumn(oB:colPos)

   // use temp for safety
   mGetVar := Eval(oCol:block)

   // create a corresponding GET with ambiguous set/get block
   oGet := GetNew(Row(), Col(),                                                                    ;
                              {|x| if(PCount() == 0, mGetVar, mGetVar := x)},      ;
                              "mGetVar",, oB:colorSpec)

   lFresh := .f.                           // refresh flag

   // read it
   if CuaRead( {oGet} )
      // new data has been entered
      if ( lAppend .and. Recno() == LastRec() + 1 )
         APPEND BLANK
      end

      // replace with new data
      Eval( oCol:block, mGetVar )

      // test for change in index order
      if ( !Empty(cExpr) .and. !lAppend )
         if ( xEval != &cExpr )
            lFresh := .t.
         end
      end
   end

   if ( lFresh )
      // record in new indexed order
      FreshOrder(oB)

      // no other action
      nKey := 0
   else
      // refresh the current row only
      oB:refreshCurrent()

      // certain keys move cursor after edit if no refresh
      nKey := ExitKey(lAppend)
   end

   if ( lAppend )
      // maintain special row color
      oB:colorRect({oB:rowPos,1,oB:rowPos,oB:colCount}, {2,2})
   end

   // restore state
   SetCursor(nCursSave)
   Set(_SET_EXIT, lExitSave)
   RtnSetProc( lRtnProc )

return (nKey)


/***
*	ExitKey()
*
*	Determine the follow-up action after editing a field
*/

static func ExitKey(lAppend)

local nKey

	nKey := LastKey()
	if ( nKey == K_PGDN )
		// move down if not append mode
		if ( lAppend )
			nKey := 0
		else
			nKey := K_DOWN
		end

	elseif ( nKey == K_PGUP )
		// move up if not append mode
		if ( lAppend )
			nKey := 0
		else
			nKey := K_UP
		end

	elseif ( nKey == K_RETURN .or. (nKey >= 32 .and. nKey <= 255) )
		// return key or type out..move right
		nKey := K_RIGHT

	elseif ( nKey != K_UP .and. nKey != K_DOWN )
		// no other action
		nKey := 0
	end

return (nKey)


/***
*	FreshOrder()
*
*	Refresh respecting any change in index order
*/

static func FreshOrder(oB)

local nRec

	nRec := Recno()
	oB:refreshAll()

	// stabilize to see if TBrowse moves the record pointer
        RtnCursor( .F. )
	while ( !oB:stabilize() ) ; end
        RtnCursor( .T. )

	if ( nRec != LastRec() + 1 )
		// record pointer may move if bof is on screen
		while ( Recno() != nRec )
			// falls through unless record is closer to bof than before
			oB:up()
                        RtnCursor( .F. )
			while ( !oB:stabilize() ) ; end
                        RtnCursor( .T. )
		end
	end

return (NIL)


/***
*	Statline()
*
*	display status at coordinates relative to TBrowse object
*/

static func Statline(oB, lAppend)

local nTop, nLeft

   nTop  := oB:nBottom + 1
   nLeft := oB:nLeft + 9

   if ( LastRec() == 0 .and. !lAppend )                    // file is empty
      @ nTop, nLeft SAY "<none>"   COLOR aClr[ CLR_DLGBOX, 4 ]

   elseif ( Recno() == LastRec() + 1 )
      @ nTop, nLeft SAY " <new>"   COLOR aClr[ CLR_DLGBOX, 4 ]
   else
      // normal record..display Recno()/LastRec() and Deleted()
      @ nTop, nLeft      SAY Pad( Ltrim(Str(Recno())) + "/" +;
                                  Ltrim(Str(LastRec())), 16 );
                         COLOR aClr[ CLR_DLGBOX, 4 ]
      @ nTop, nLeft + 31 SAY If( Deleted(), cChkON, cChkOFF ) + "Deleted";
                         COLOR aClr[ CLR_DLGBOX, 4 ]
      @ nTop, nLeft + 52 SAY If( oB:hitTop, cChkON, cChkOFF ) + "B.O.F.";
                         COLOR aClr[ CLR_DLGBOX, 4 ]
   end

return (NIL)


/***
*	Skipped(n)
*
*	Skip thru database and return the
*	actual number of records skipped
*/

static func Skipped(nRequest, lAppend)

local nCount

	nCount := 0
	if ( LastRec() != 0 )
		if ( nRequest == 0 )
			skip 0

		elseif ( nRequest > 0 .and. Recno() != LastRec() + 1 )
			// forward
			while ( nCount < nRequest )
				skip 1
				if ( Eof() )
					if ( lAppend )
						// eof record allowed if append mode
						nCount++
					else
						// back to last actual record
						skip -1
					end

					exit
				end

				nCount++
			end

		elseif ( nRequest < 0 )
			// backward
			while ( nCount > nRequest )
				skip -1
				if ( Bof() )
					exit
				end

				nCount--
			end
		end
	end

return (nCount)


// eof _browse.prg
