/*-----------------------------------------------------------------------------

                   Improved GetSys to manage HotKeys and Mouse


   (c) Francisco Morero Peyrona - 1993
       Alfonso Fraguas Bravo    - 1993

-----------------------------------------------------------------------------*/
#include "Set.ch"
#include "Inkey.ch"
#include "Getexit.ch"
#include "TDCO.ch"

//--------------------------------------------- state variables for active READ
static Format
static Updated := .f.
static KillRead
static BumpTop
static BumpBot
static LastExit
static LastPos
static ActiveGet
static ReadProcName
static ReadProcLine
static aRtn	   := { 0, 0, 0 }
static lClick      := .F.                            // mouse or keyboard event?

//----------------------------- format of array used to preserve state variables
#define GSV_KILLREAD		1
#define GSV_BUMPTOP		2
#define GSV_BUMPBOT		3
#define GSV_LASTEXIT		4
#define GSV_LASTPOS		5
#define GSV_ACTIVEGET		6
#define GSV_READVAR		7
#define GSV_READPROCNAME	8
#define GSV_READPROCLINE	9
#define GSV_COUNT		9

/*------------------------------------------------------------------------------
  This function is here for compatibility with previous code only.
  If you don't need it, remove it.
------------------------------------------------------------------------------*/
function ReadModal( aGetList )
return CuaRead( aGetList )


/*----------------------------------------------------------------------------//
   Esta funcion pone ne marcha un aGetList, llamando a CuaRead() y
   activa K_CTRL_DOWN o no, dependiendo de si algun ComboBox o no en el
   aGetList. Tambien crea una array en oGet:cargo para los gets normales
   de Clipper con los elementos de los objetos CuaWise siguientes

   [ CUA_CLASS ] == "CW_NOCLASS" -> Nombre de la clase
   [ CUA_AREA  ] == {}           -> Area ocupada por ese objeto

   El resto de los items del array almacenado en oGet:Cargo, estn vacos.
   Aqu puede almacenarse lo que se necesite (en tiempo de ejecucin, claro).

------------------------------------------------------------------------------*/
function CuaRead( aGetList )

   local i
   local nLen   := Len( aGetList )
   local nCrs   := SetCursor( If( ReadInsert(), 2, 1 ) )
   local lCrs   := RtnCursor()
   local bKeyCmb:= SetKey( K_CTRL_DOWN )
   local lCombo := .F.                         // Hay algun combo en el GetList?

   for i = 1 to nLen
       if aGetList[ i ]:cargo == NIL
          aGetList[ i ]:cargo = CargoStandar( aGetList[ i ] )

       elseif aGetList[ i ]:cargo[ CUA_CLASS ] == "CW_COMBO"
          lCombo = .T.
       endif
   next

   if lCombo
      SetKey( K_CTRL_DOWN, { || ComboCua( GetActive() ) } )
   endif

   CuaModal( aGetList )                                // La madre del cordero

   RtnCursor( lCrs )
   SetKey( K_CTRL_DOWN, bKeyCmb )
   SetCursor( nCrs )

return ( Updated )

//--------------------------------------------------//

static func CargoStandar( oGet )

   local aCargo := Array( CUA_LEN ), nAncho

   aCargo[ CUA_CLASS ] = "CW_NOCLASS"

   oGet:SetFocus()  ;  nAncho = Len( oGet:Buffer ) - 1  ;  oGet:KillFocus()

   if !Empty( oGet:picture )
      nAncho = Len( Transform( Space( nAncho ), oGet:Picture ) )
   endif

   aCargo[ CUA_AREA ] = { oGet:Row, oGet:Col, oGet:Row, oGet:Col+nAncho }
   aCargo[ CUA_HOTKEY ] = { 0, 0, "", 0 }

return aCargo


/***
*       CuaModal()
*	Standard modal READ on an array of GETs.
*/
static func CuaModal( GetList )

local nPos, savedGetSysVars, oGet


   if ( ValType(Format) == "B" )
      Eval(Format)
   end

   aRtn 	   = { 0, 0, 0 }
   lClick          = .F.
   savedGetSysVars = ClearGetSysVars()
   ReadProcName    = ProcName( 1 )
   ReadProcLine    = ProcLine( 1 )

   nPos := Settle( Getlist, 0 )

   while ( nPos != 0 )
      oGet = GetList[ nPos ]
      PostActiveGet( oGet )

      if oGet:reader != NIL
	 Eval( oGet:reader, { GetList, nPos } )
      else
	 GetReader( GetList, nPos )
      endif

      nPos = Settle( GetList, nPos )
   enddo

   RestoreGetSysVars( savedGetSysVars )

return ( Updated )


/***
*	Reader() para los Gets normales
*/

static proc GetReader( GetList, nGet )

local oGet := GetList[ nGet ]

   if ( GetPreValidate( oGet ) )
      oGet:SetFocus()

      do while ( oGet:exitState == GE_NOEXIT )
	 if ( oGet:typeOut )
	    oGet:exitState := GE_ENTER
	 endif

	 do while ( oGet:exitState == GE_NOEXIT )
	    GetApplyKey( oGet, CuaKeyRtn( GetList, nGet ) )
	 enddo

	 if ( !GetPostValidate(oGet) )
	    oGet:exitState := GE_NOEXIT
	 endif
      enddo

      oGet:KillFocus()
   endif

return

/*------------------------------------------------------
   Toma una tecla o una pulsacion de raton

------------------------------------------------------*/
func CuaKeyRtn( GetList, nGet )

   local nPos := 0, nRet := 0, nTime
   local nDark := aCfg[ CFG_SCR_DARK ] * 60

   RtnCursor( .T. )

   do while .T.
      // hay que inicializar estos valores
      aRtn[ nRTN_B ] = 0 ; lClick = .F. ; nTime = Seconds() 

      do while aRtn[ nRTN_B ] != 1 .and. NextKey() == 0
	 aRtn = RtnEstado()
         if Seconds() - nTime > nDark
            Cielo()  ;  nTime = Seconds()
         endif
      enddo

      if NextKey() > 271                        //  Se ha pulsado un HotKey ?
         AplicaHotKey( GetList, NextKey() )     // Cambia el contenido de aRtn[]
      endif

      if (aRtn[ nRTN_B ] == 1)
         if (nPos := CuaBuscaArea( GetList )) != 0
            lClick = .T.
            if nPos == nGet
               AplicaClick( GetList[ nPos ] )      // Deja tecla en el buffer
            else
               GetList[ nGet ]:ExitState = 9       // Pasar focus a otro obj
               nRet = 0  ;  EXIT
            endif
         endif
      endif

      if NextKey() != 0
	 nRet = Inkey() ; EXIT
      endif
   enddo

   RtnCursor( .F. )

return nRet


/*------------------------------------------------------------------------------
   Busca si la tecla pulsada corresponde a algun HotKey,
   de ser asi, altera el contenido de aRtn[] para simular una pulsacion
   del raton sobre la posicion del HotKey

-----------------------------------------------------------------------*/
static func AplicaHotKey( aGetLst, nKey )

   local i

   for i = 1 to Len( aGetLst )
       if Left( aGetLst[ i ]:cargo[ CUA_CLASS ], 3 ) == "CW_"
          if aGetLst[ i ]:cargo[ CUA_HOTKEY, 4 ] == nKey
             aRtn[ nRTN_Y ] = aGetLst[ i ]:cargo[ CUA_HOTKEY, 1 ]
             aRtn[ nRTN_X ] = aGetLst[ i ]:cargo[ CUA_HOTKEY, 2 ]
             aRtn[ nRTN_B ] = 1
          endif
       endif
   next

return NIL


/*------------------------------------------------------------------------------
   De una tecla en el buffer, es decir dependiendo del tipo de
   objeto GET del q se trate y de si este ya tenia o no el Focus

-----------------------------------------------------------------------*/
static func AplicaClick( oGet )

   local xAux

   do case
      case oGet:cargo[ CUA_CLASS ] == "CW_GET" .or.;
           oGet:cargo[ CUA_CLASS ] == "CW_NOCLASS"

           KEYBOARD Chr( K_HOME ) + Replicate( Chr( K_RIGHT ),;
                                               aRtn[ nRTN_X ] - oGet:Col )

      case oGet:cargo[ CUA_CLASS ] == "CW_BUTTON"
	   KEYBOARD Chr( 32 )

      case oGet:cargo[ CUA_CLASS ] == "CW_RADIO"
           if aRtn[ nRTN_Y ] == oGet:cargo[ CUA_AREA, 1 ];
             .and. oGet:cargo[ CUA_LABEL ] != ""              // Label
                   // Si se clickea sobre el label, no tomar accion
           else
              xAux = aRtn[ nRTN_Y ] - oGet:cargo[ CUA_DIM, 1 ] + 1
              Eval( oGet:Cargo[ CUA_BLOCK ], xAux )
              oGet:Cargo[ CUA_CARGO, 1 ] = xAux
           endif
           KEYBOARD Chr( K_PGUP )                             // * devolver algo

      case oGet:cargo[ CUA_CLASS ] == "CW_CHECK"
           if aRtn[ nRTN_Y ] == oGet:cargo[ CUA_AREA, 1 ];
             .and. oGet:cargo[ CUA_LABEL ] != ""              // Label
              KEYBOARD Chr( K_PGUP )                          // * devolver algo
           else
              oGet:cargo[ CUA_LEN + 1 ] = aRtn[ nRTN_Y ] -;
                                          oGet:cargo[ CUA_DIM, 1 ] + 1
              KEYBOARD Chr( 32 )  ;  Sleep( .18 )
           endif

      case oGet:cargo[ CUA_CLASS ] == "CW_COMBO"
           if aRtn[ nRTN_X ] == oGet:cargo[ CUA_AREA, 4 ]     // sobre la flecha
	      KeyBuff( KB_CTRL_DOWN )
	   else
              KEYBOARD Chr( K_HOME ) + Replicate( Chr( K_RIGHT ),;
                                                  aRtn[ nRTN_X ] - oGet:Col )
	   endif

      case oGet:cargo[ CUA_CLASS ] == "CW_LISTBOX"
           if aRtn[ nRTN_Y ] == oGet:cargo[ CUA_AREA, 1 ] .and.;      // ListBox
              aRtn[ nRTN_X ] == oGet:cargo[ CUA_AREA, 2 ] .and.;      // de un
              oGet:picture == "LIST-COMBO"                            // combo
              KEYBOARD Chr( K_ESC )

	   elseif aRtn[ nRTN_X ] == oGet:cargo[ CUA_AREA, 4 ]
              if aRtn [ nRTN_Y ] == oGet:Row + 1                   // Up arrow
		  KEYBOARD Chr( K_UP )
              elseif aRtn[ nRTN_Y ] == oGet:cargo[ CUA_AREA, 3 ]-1 // Down arrow
		  KEYBOARD Chr( K_DOWN )
	      endif

           elseif aRtn[ nRTN_Y ] > oGet:Row .and.;                    // Item
                  aRtn[ nRTN_Y ] < oGet:cargo[ CUA_AREA, 3 ] .and.;
		  aRtn[ nRTN_X ] > oGet:Col .and.;
                  aRtn[ nRTN_X ] < oGet:cargo[ CUA_AREA, 4 ]

		  xAux = ( aRtn[ nRTN_Y ] - oGet:Row ) - 1 -;
		       ( oGet:cargo[ CUA_CARGO, 1 ] - oGet:cargo[ CUA_DIM, 3 ] )

		  if xAux == 0
		     KEYBOARD Chr( K_ENTER )
		  else
		     xAux = Replicate( Chr( If(xAux>0,K_DOWN,K_UP) ), Abs( xAux ) )
		     KEYBOARD xAux
		  endif
	   endif
   endcase

return NIL

//----------------------------------------------------------------------------//

func CuaBuscaArea( GetList )

   local i := 1

   for i = 1 to Len( GetList )
       if Left( GetList[ i ]:cargo[ CUA_CLASS ], 3 ) == "CW_"
	  if aRtn[ nRTN_Y ] >= GetList[ i ]:Cargo[ CUA_AREA, 1 ] .and.;
	     aRtn[ nRTN_X ] >= GetList[ i ]:Cargo[ CUA_AREA, 2 ] .and.;
	     aRtn[ nRTN_Y ] <= GetList[ i ]:Cargo[ CUA_AREA, 3 ] .and.;
	     aRtn[ nRTN_X ] <= GetList[ i ]:Cargo[ CUA_AREA, 4 ]
	     return i
	  endif
       endif
   next

return 0

//----------------------------------------------------------------------------//

/***
*	GetApplyKey()
*	Apply a single Inkey() keystroke to a GET.
*
*	NOTE: GET must have focus.
*/
proc GetApplyKey(get, key)

local cKey
local bKeyBlock


     if ( (bKeyBlock := SetKey(key)) <> NIL )
	GetDoSetKey( bKeyBlock, get )
	return									// NOTE
     end

     do case
	case ( key == K_ESC )
	     if ( Set(_SET_ESCAPE) )
		get:undo()
		get:exitState := GE_ESCAPE
	     end

	case ( key == K_UP )	     ;	get:exitState := GE_UP
	case ( key == K_SH_TAB )     ;	get:exitState := GE_UP
	case ( key == K_DOWN )	     ;	get:exitState := GE_DOWN
	case ( key == K_TAB )	     ;	get:exitState := GE_DOWN
	case ( key == K_ENTER )      ;	get:exitState := GE_ENTER
	case ( key == K_PGUP )	     ;	get:exitState := GE_WRITE
	case ( key == K_PGDN )	     ;	get:exitState := GE_WRITE
	case ( key == K_CTRL_HOME )  ;	get:exitState := GE_TOP
	case ( key == K_CTRL_W )     ;	get:exitState := GE_WRITE
	case ( key == K_INS )	     ;	Set( _SET_INSERT, !Set(_SET_INSERT) )
	case ( key == K_CTRL_U )     ;	get:Undo()
	case ( key == K_HOME )	     ;	get:Home()
	case ( key == K_END )	     ;	get:End()
	case ( key == K_RIGHT )      ;	get:Right()
	case ( key == K_LEFT )	     ;	get:Left()
	case ( key == K_CTRL_RIGHT ) ;	get:WordRight()
	case ( key == K_CTRL_LEFT )  ;	get:WordLeft()
	case ( key == K_BS )	     ;	get:BackSpace()
	case ( key == K_DEL )	     ;	get:Delete()
	case ( key == K_CTRL_T )     ;	get:DelWordRight()
	case ( key == K_CTRL_Y )     ;	get:DelEnd()
	case ( key == K_CTRL_BS )    ;	get:DelWordLeft()

     otherwise

	if (key >= 32 .and. key <= 255)
	   cKey := Chr(key)

	   if (get:type == "N" .and. (cKey == "." .or. cKey == ","))
		   get:ToDecPos()
	   else
	      if ( Set( _SET_INSERT ) )
		 get:Insert( cKey )
	      else
		 get:Overstrike( cKey )
	      endif
	      if (get:typeOut .and. !Set(_SET_CONFIRM) )
		 if ( Set(_SET_BELL) )
		      ?? Chr(7)
		 end
		 get:exitState := GE_ENTER
	      endif
	   endif
	endif

     endcase

return

//----------------------------------------------------------------------------//
/***
*	GetDoSetKey()
*	Process SET KEY during editing.
*/
proc GetDoSetKey(keyBlock, get)

local saveUpdated

   // if editing has occurred, assign variable
   if ( get:changed )
       get:Assign()
       Updated := .t.
   end

   saveUpdated := Updated
   Eval( keyBlock, ReadProcName, ReadProcLine, ReadVar() )
   get:UpdateBuffer()
   Updated := saveUpdated

   if ( KillRead )
      get:exitState := GE_ESCAPE              // provokes CuaRead() exit
   end

return

//----------------------------------------------------------------------------//

/***
*	Settle() : Returns new position in array of Get objects, based on
*
*		- current position
*		- exitState of Get object at current position
*
*	NOTE return value of 0 indicates termination of READ
*	NOTE exitState of old Get is transferred to new Get
*/
static func Settle( GetList, nPos )

local exitState, nNewPos

   if ( nPos == 0 )
      exitState := GE_DOWN
   else
      exitState := GetList[ nPos ]:exitState
   end

   if ( exitState == GE_ESCAPE .or. exitState == GE_WRITE )
      return ( 0 )				      // NOTE
   end

   if ( exitState <> GE_WHEN )
      // reset state info
      LastPos := nPos
      BumpTop := .f.
      BumpBot := .f.
   else
      // re-use last exitState, do not disturb state info
      exitState := LastExit
   end

   /***
   *	   move
   */
   if lClick
      if ( nNewPos := CuaBuscaArea( GetList ) ) != 0
	 if GetList[ nNewPos ]:PreBlock == NIL
	    nPos = nNewPos
	    AplicaClick( GetList[ nPos ] )

	 elseif Eval( GetList[ nNewPos ]:PreBlock, GetList[ nNewPos ] )

	    nPos = nNewPos
	    AplicaClick( GetList[ nPos ] )
	 endif
     endif

   else
						       // pulsado tecla
      do case
         case ( exitState == GE_UP )    ;  nPos--
         case ( exitState == GE_DOWN )  ;  nPos++
	 case ( exitState == GE_TOP )
	      nPos := 1
	      BumpTop := .T.
	      exitState := GE_DOWN

	 case ( exitState == GE_BOTTOM )
	      nPos := Len(GetList)
	      BumpBot := .T.
	      exitState := GE_UP

	 case ( exitState == GE_ENTER ) ; nPos ++
      endcase
   endif

   /***
   *	   bounce
   */
   if ( nPos == 0 )                                            // bumped top
      if ( !ReadExit() .and. !BumpBot )
          BumpTop   := .T.
          nPos      := LastPos
	  exitState := GE_DOWN
      end

   elseif ( nPos == Len(GetList) + 1 )                         // bumped bottom

      if ( !ReadExit() .and. exitState <> GE_ENTER .and. !BumpTop )
         BumpBot   := .T.
         nPos      := LastPos
	 exitState := GE_UP
      else
	 nPos := 0
      endif
   endif

   // record exit state
   LastExit := exitState

   if ( nPos <> 0 )
      GetList[ nPos ]:exitState := exitState
   end

return (nPos)


/***
*	Post active GET for ReadVar(), GetActive().
*/
static proc PostActiveGet(get)

   GetActive( get )
   ReadVar( GetReadVar(get) )

return


/***
*	Save and clear READ state variables. Return array of saved values.
*
*	NOTE: 'Updated' status is cleared but not saved (S87 compat.).
*/
static func ClearGetSysVars()

local saved[ GSV_COUNT ]


	saved[ GSV_KILLREAD ] := KillRead
	KillRead := .f.

	saved[ GSV_BUMPTOP ] := BumpTop
	BumpTop := .f.

	saved[ GSV_BUMPBOT ] := BumpBot
	BumpBot := .f.

	saved[ GSV_LASTEXIT ] := LastExit
	LastExit := 0

	saved[ GSV_LASTPOS ] := LastPos
	LastPos := 0

	saved[ GSV_ACTIVEGET ] := GetActive( NIL )

	saved[ GSV_READVAR ] := ReadVar( "" )

	saved[ GSV_READPROCNAME ] := ReadProcName
	ReadProcName := ""

	saved[ GSV_READPROCLINE ] := ReadProcLine
	ReadProcLine := 0

	Updated := .f.

return (saved)


/***
*	Restore READ state variables from array of saved values.
*
*	NOTE: 'Updated' status is not restored (S87 compat.).
*/
static proc RestoreGetSysVars(saved)

	KillRead := saved[ GSV_KILLREAD ]

	BumpTop := saved[ GSV_BUMPTOP ]

	BumpBot := saved[ GSV_BUMPBOT ]

	LastExit := saved[ GSV_LASTEXIT ]

	LastPos := saved[ GSV_LASTPOS ]

	GetActive( saved[ GSV_ACTIVEGET ] )

	ReadVar( saved[ GSV_READVAR ] )

	ReadProcName := saved[ GSV_READPROCNAME ]

	ReadProcLine := saved[ GSV_READPROCLINE ]

return


/***
*	Set READVAR() value from a GET.
*/
static func GetReadVar( get )

local name := Upper( get:name ), i

   if ( get:subscript <> NIL )
      for i := 1 to Len( get:subscript )
	  name += "[" + LTrim( Str( get:subscript[ i ] ) ) + "]"
      next
   endif

return (name)


/*************************************
*
*	system services
*
*/

/***
*   __SetFormat()
*	SET FORMAT service
*/
func __SetFormat(b)
	Format := if ( ValType(b) == "B", b, NIL )
return (NIL)


/***
*	__KillRead()
*   CLEAR GETS service
*/
proc __KillRead()
	KillRead := .t.
return


/***
*	GetActive()
*/
func GetActive(g)
local oldActive := ActiveGet
	if ( PCount() > 0 )
		ActiveGet := g
	end
return ( oldActive )


/***
*	Updated()
*/
func Updated()
return (Updated)


/***
*	ReadExit()
*/
func ReadExit(lNew)
return ( Set(_SET_EXIT, lNew) )

/***
*	ReadInsert()
*/
func ReadInsert(lNew)
return ( Set(_SET_INSERT, lNew) )

//----------------------------------------------------------------------------//

func GetPreValidate(get)

local saveUpdated
local when := .t.

   if ( get:preBlock <> NIL )
      saveUpdated := Updated
      when := Eval(get:preBlock, get)
      get:Display()
      Updated := saveUpdated
   end

   if ( KillRead )
      when := .f.
      get:exitState := GE_ESCAPE	      // provokes ReadModal() exit
   elseif ( !when )
      get:exitState := GE_WHEN		      // indicates failure
   else
      get:exitState := GE_NOEXIT	      // prepares for editing
   end

return (when)


/***
*	GetPostValidate()
*	Test exit condition (VALID clause) for a GET.
*
*	NOTE: bad dates are rejected in such a way as to preserve edit buffer.
*/
func GetPostValidate(get)

local saveUpdated
local changed, valid := .t.


	if ( get:exitState == GE_ESCAPE )
	   return (.t.) 				   // NOTE
	end

	if ( get:BadDate() )
	   get:Home()
	   return (.f.) 				   // NOTE
	end

	// if editing occurred, assign the new value to the variable
	if ( get:changed )
	    get:Assign()
	    Updated := .t.
	end

	// reform edit buffer, set cursor to home position, redisplay
	get:Reset()

	// check VALID condition if specified
	if ( get:postBlock <> NIL )
	   saveUpdated := Updated
	   valid := Eval( get:postBlock, get )

	   // reset compat. pos
	   SetPos( get:row, get:col )

	   get:UpdateBuffer()
	   Updated := saveUpdated

	   if ( KillRead )
	      get:exitState := GE_ESCAPE      // provokes ReadModal() exit
	      valid := .t.
	   endif
	endif

return (valid)
